unit uLemmingsGame;

interface

uses FastGL, GLHelper, dglOpenGL, glBitmap, Geometry, Classes, Windows, Controls,
     Graphics, ExtCtrls, SysUtils, uResources;

type
  TLemmingsWorld = class;

  TLemmingsObject = class
  private
    FWorld:TLemmingsWorld;
  public
    constructor Create(AWorld:TLemmingsWorld); virtual;
    property World : TLemmingsWorld read FWorld write FWorld;
    procedure Render; virtual;
  end;

  TLemmingsObjectList = class(TList)
  private
    function Get(Index: Integer): TLemmingsObject;
    procedure Put(Index: Integer; const Value: TLemmingsObject);
  public
    function Add(Item: TLemmingsObject): Integer;
    function IndexOf(Item: TLemmingsObject): Integer;
    procedure Insert(Index: Integer; Item: TLemmingsObject);
    function Remove(Item: TLemmingsObject): Integer;
    property Items[Index: Integer]: TLemmingsObject read Get write Put; default;
  end;    

  TLemmingsVisObject  = class(TLemmingsObject)
  private
    FTexture: TglBitmap2D;
    FOffs: TPoint;
    FBB: TRect;
  protected
    FPos: TVector3f;
    function WorldBBox:TRect;
  public
    property Position:TVector3f read FPos   write FPos;
    property Offset  :TPoint   read FOffs  write FOffs;

    property BoundingBox:TRect read FBB    write FBB;
    property Texture :TglBitmap2D read FTexture;
    procedure OpenTexture(FN:string; OffsX,OffsY:integer;BBox:TRect);

    procedure Render; override;

    constructor Create(AWorld:TLemmingsWorld); override;
    destructor Destroy; override;
  end;

  TCollisionResult = (crEmpty,crBadField,crGoodField);
  TLemmingsMap = class(TLemmingsObject)
  private
    Level:TglBitmap2D;
    alpha:TBitmap;
  protected
  public
    constructor Create(AWorld:TLemmingsWorld); override;
    destructor Destroy; override;

    procedure LoadFromFile(FN:string);

    procedure ExplosionAt(X,Y,R:TGLfloat);
    procedure Render; override;

    function CheckForCollision(Rect: TRect):TCollisionResult;
  end;

  TControlButton = (cbNone,cbBlocker,cbDigHorz,cbDigVert,cbExplode);
  TControlButtonInf = record
                        pos:TRect;
                        Count:integer;
                      end;
  TLemmingsControlbar = class(TLemmingsVisObject)
  private
    FCurrentTool: TControlButton;
  protected
    Buttons:array[TControlButton] of TControlButtonInf;
    function FindButton(pt:TPoint):TControlButton;
  public
    constructor Create(AWorld:TLemmingsWorld); override;
    procedure HandleMouseClick(pt: TPoint);
    property CurrentTool : TControlButton read FCurrentTool write FCurrentTool; 
  end;


  TRenderState = (rsNone, rsLevelSelect, rsGame);
  TLemmingsWorld = class
  private
    FMap: TLemmingsMap;
    FControlBar:TLemmingsControlbar;
    FWND:TWinControl;
    FSubObjs: TLemmingsObjectList;
    FState: TRenderState;
    FViewXY:TPoint;
  protected
    DC:HDC;
    RC:HGLRC;

    Fnt_Arial_huge,
    Fnt_Arial_normal,
    Fnt_Arial_normal_bold:TGLuint;
    LevelList:TStringList;
    LevelIndex:integer;
    LevelPrev:TglBitmap2D;
    LevelCreator,LevelTitle:string;
    procedure SetLevelIndex(new:integer);
  public
    constructor Create;
    destructor Destroy; override;
    procedure RenderInit(Window:TWinControl);

    property State : TRenderState read FState;
    function ChangeState(AState:TRenderState):boolean;
    procedure Render;
    procedure RenderGame;
    procedure RenderLevelSelect;
    function GetTimer:DWORD;

    property Map : TLemmingsMap read FMap;
    property ViewXY : TPoint read FViewXY write FViewXY; 
    property SubObjects : TLemmingsObjectList read FSubObjs;
    function ClientToScreen(pt:TPoint):TPoint;
    procedure HandleKeys;
    procedure HandleMouseClick(pt:TPoint);

    procedure LoadLevel(fn:string);
  end;

implementation

uses Math,DUStrings, DUXMLParser, uLemmingsObjects, StdConvs;

Procedure FindFiles(Path:String;Attr:integer;Strings:TStrings;SubDirs:boolean);
var SRec:TSearchrec;
    error:integer;
begin
 //strings.clear;
 if Subdirs
  then begin
   error:=findFirst(path,attr or faDirectory,SRec);
   while error=0 do
    begin
     if (Srec.Attr and faDirectory)<>0then FindFiles(Path,Attr,Strings,Subdirs);
     if (SRec.Attr or Attr)=Attr
      then strings.add(extractfilepath(Path)+srec.Name);
     error:=findnext(SRec);
    end;
   findclose(SRec);
  end
  else begin
   error:=findFirst(path,attr,SRec);
   while error=0 do
    begin
     strings.add(srec.Name);
     error:=findnext(SRec);
    end;
   findclose(SRec);
  end;
end;

{ TLemmingsObject }

constructor TLemmingsObject.Create(AWorld: TLemmingsWorld);
begin
  inherited Create;
  FWorld:= AWorld;
end;

procedure TLemmingsObject.Render;
begin
end;

{ TLemmingsObjectList }

function TLemmingsObjectList.Add(Item: TLemmingsObject): Integer;
begin
  Result:= inherited Add(Item);
end;

function TLemmingsObjectList.Get(Index: Integer): TLemmingsObject;
begin
  Result:=  inherited Get(Index);
end;

function TLemmingsObjectList.IndexOf(Item: TLemmingsObject): Integer;
begin
  Result:= inherited IndexOf(Item);
end;

procedure TLemmingsObjectList.Insert(Index: Integer; Item: TLemmingsObject);
begin
  inherited Insert(Index,Item);
end;

procedure TLemmingsObjectList.Put(Index: Integer; const Value: TLemmingsObject);
begin
  inherited Put(Index,Value);
end;

function TLemmingsObjectList.Remove(Item: TLemmingsObject): Integer;
begin
  Result:= inherited Remove(Item);
end;

{ TLemmingsVisObject }

constructor TLemmingsVisObject.Create(AWorld: TLemmingsWorld);
begin
  inherited;
  FTexture:= TglBitmap2D.Create;
  FTexture.FreeDataAfterGenTexture:= false;
  World.SubObjects.Add(Self);
end;

destructor TLemmingsVisObject.Destroy;
begin
  World.SubObjects.Remove(Self);
  FTexture.Free;
  inherited;
end;

procedure TLemmingsVisObject.OpenTexture(FN: string; OffsX, OffsY: integer;
  BBox: TRect);
begin
  FTexture.LoadFromFile(FN);
  FTexture.Target:= GL_TEXTURE_RECTANGLE_EXT;
  FTexture.GenTexture;
  FOffs:= point(OffsX,OffsY);
  FBB:= BBox;
end;

procedure TLemmingsVisObject.Render;
begin
  inherited;
  FTexture.Bind;
  glPushMatrix;
    glTranslatef(FPos[0],FPos[1],0);
    glTranslatef(-FOffs.X,-FOffs.Y,0);
    glBegin(GL_QUADS);
      glTexCoord2f(0,0);                           glVertex2d(0  ,0);
      glTexCoord2f(FTexture.Width,0);              glVertex2d(FTexture.Width,0);
      glTexCoord2f(FTexture.Width,FTexture.Height);glVertex2d(FTexture.Width,FTexture.Height);
      glTexCoord2f(0,FTexture.Height);             glVertex2d(0  ,FTexture.Height);
    glEnd;
  glPopMatrix;
end;

function TLemmingsVisObject.WorldBBox: TRect;
begin
  Result:= TranslateRect(FBB,-FOffs.X+1,-FOffs.Y+1);
  Result:= TranslateRect(Result,trunc(FPos[0]),trunc(FPos[1]));
end;

{ TLemmingsMap }

constructor TLemmingsMap.Create(AWorld: TLemmingsWorld);
begin
  inherited;
  Level:= TglBitmap2D.Create;
  alpha:= TBitmap.Create;
  Level.FreeDataAfterGenTexture:= false;
  Level.Target:= GL_TEXTURE_RECTANGLE_EXT;
  Level.SetFilter(GL_LINEAR,GL_LINEAR);
  Level.Format:= tfCompressed;
end;

destructor TLemmingsMap.Destroy;
begin
  Level.Free;
  alpha.Free;
end;

procedure TLemmingsMap.LoadFromFile(FN: string);
begin
  Level.LoadFromFile(FN);
  Level.AssignAlphaToBitmap(alpha);
  Level.GenTexture;
end;

procedure TLemmingsMap.ExplosionAt(X, Y, R: TGLfloat);
var i,j:integer;
    line:PByteArray;
    d:TGLFloat;
begin
  for i:= 1 to alpha.Height do begin
   line:= alpha.Scanline[i-1];
    for j:= 1 to alpha.Width do
      if line^[j-1]=mapGood then begin
        d:= sqrt(sqr(X-j)+sqr(y-i));
        if d<=r then
          line^[j-1]:= mapEmpty;
      end;
  end;
  Level.AddAlphaFromBitmap(alpha);
  Level.GenTexture;
end;

procedure TLemmingsMap.Render;
begin
  inherited;
  Level.Bind;
  glBegin(GL_QUADS);
    glTexCoord2f(0,Level.Height);
      glVertex2d(0  ,Level.Height);
    glTexCoord2f(Level.Width,Level.Height);
      glVertex2d(Level.Width,Level.Height);
    glTexCoord2f(Level.Width,0);
      glVertex2d(Level.Width,0);
    glTexCoord2f(0,0);
      glVertex2d(0  ,0);
  glEnd;
end;



function TLemmingsMap.CheckForCollision(Rect: TRect):TCollisionResult;
var i,j:integer;
    line:PByteArray;
begin
  for i:= Rect.Top to Rect.Bottom do begin
    if (i<=0) or (i>alpha.Height) then continue;
    line:= alpha.Scanline[i-1];
    for j:= Rect.Left to Rect.Right do begin
      if (j<=0) or (j>alpha.Width) then continue;
      case line^[j-1] of
        mapGood:begin Result:= crGoodField; exit end;
        mapBad:begin Result:= crBadField; exit end;
      end;
    end;
  end;
  Result:= crEmpty;
end;

{ TLemmingsWorld }

constructor TLemmingsWorld.Create;
begin
  inherited;
  FMap:= TLemmingsMap.Create(Self);
  FSubObjs:= TLemmingsObjectList.Create;
  FWND:= nil;
  LevelList:= TStringList.Create;
  LevelPrev:= TglBitmap2D.Create;
  LevelPrev.Format:= tfCompressed;

  FState:= rsNone;
  ChangeState(rsLevelSelect);
end;

destructor TLemmingsWorld.Destroy;
begin
  glCloseRender(RC);
  FSubObjs.Free;
  FMap.Free;
  LevelList.Free;
  LevelPrev.Free;
end;

function TLemmingsWorld.ClientToScreen(pt: TPoint): TPoint;
begin
  Result.X:= Round(pt.X/FWND.ClientWidth*FMap.Level.Width);
  Result.Y:= Round(pt.Y/FWND.ClientHeight*FMap.Level.Height);
end;

procedure TLemmingsWorld.RenderInit(Window: TWinControl);
begin
  FWND:= Window;
  glOpenRender(Window.Handle,Window.ClientWidth,Window.ClientHeight,DC,RC);

  Fnt_Arial_huge:= glInitFont('arial',DC,50);
  Fnt_Arial_normal:= glInitFont('arial',DC,20);
  Fnt_Arial_normal_bold:= glInitFont('arial',DC,20,[fsBold]);
  LevelIndex:= -1;
  SetLevelIndex(0);

  FControlBar:= TLemmingsControlbar.Create(Self);
end;


procedure TLemmingsWorld.LoadLevel(fn: string);
var x:TXMLFile;
    s:string;
begin
  x:= TXMLFile.Create;
  try
    x.LoadFromFile(fn);

    Map.LoadFromFile('resource\level\'+x.GetText('level/file'));

    with TLemmingsDropDoor.Create(Self) do begin
      s:= x.GetAttribute('level/door','position');
      delete(s,1,1);
      Position:= GLHelper.MakeVector(StrToInt(FetchToken(s,',')),StrToInt(FetchToken(s,')')),0);
      DropCount:= StrToInt(x.GetAttribute('level/door','count'));
      DropTimer:= StrToInt(x.GetAttribute('level/door','time'));
    end;
  finally
    x.Free;
  end;
end;

function TLemmingsWorld.ChangeState(AState: TRenderState): boolean;
begin
  Result:= FState<>AState;
  if Result then begin
    FState:= AState;
    case  FState of
      rsLevelSelect:begin
                      FindFiles('resource\level\*.xml',faAnyFile,LevelList,false);
                      LevelIndex:= 0;
                    end;
      rsGame : begin

               end;
    end;
  end;
end;

procedure TLemmingsWorld.Render;
begin
  glPreDraw;
    //Licht aus
    glDisable(GL_LIGHTING);
    //Texturen an
    glEnable(GL_TEXTURE_2D);
    //los gehts
    Enter2dMode(ScreenWidth,ScreenHeight);
      case FState of
        rsLevelSelect: RenderLevelSelect;
        rsGame       : RenderGame;
      end;
    Exit2dMode;
  glAfterDraw(DC);
end;

procedure TLemmingsWorld.RenderLevelSelect;
var i:integer;
begin
  SetGLColor(c_Red);
  glRasterPos2f(10,40);
    glTextOut(Fnt_Arial_huge,SSelectLevel);

  for i:= 0 to LevelList.Count-1 do begin
    if i=LevelIndex then SetGLColor(c_White) else SetGLColor(c_Green);
    glRasterPos2f(20,60+i*20);
      glTextOut(Fnt_Arial_normal,LevelList[i]);
  end;

  SetGLColor(c_White);
  LevelPrev.Bind;
  glBegin(GL_QUADS);
    glTexCoord2f(0,LevelPrev.Height);
      glVertex2d(300  ,200);
    glTexCoord2f(LevelPrev.Width,LevelPrev.Height);
      glVertex2d(600,200);
    glTexCoord2f(LevelPrev.Width,0);
      glVertex2d(600,20);
    glTexCoord2f(0,0);
      glVertex2d(300  ,20);
  glEnd;
  LevelPrev.Unbind;

  SetGLColor(c_White);
  glRasterPos2f(300,220);
    glTextOut(Fnt_Arial_normal_bold,LevelTitle);
  glRasterPos2f(300,260);
    glTextOut(Fnt_Arial_normal,LevelCreator);
  glRasterPos2f(300,450);
    glTextOut(Fnt_Arial_normal,SUpDownSelect);
end;

procedure TLemmingsWorld.RenderGame;
var i:integer;
begin
  //Alpha-Kanal
  glEnable(GL_ALPHA_TEST);
  glAlphaFunc(GL_GREATER, 0.1);
  glPushMatrix;
    glTranslated(-FViewXY.X,-FViewXY.Y,0);
    FMap.Render;
    for i:= 0 to FSubObjs.Count-1 do
      FSubObjs[i].Render;
  glPopMatrix;
  FControlBar.Render;
end;

function KeyIsPressed(const Key: Byte): Boolean;
begin
  Result := Key and $80 <> 0; // checks if high-order bit is set, i.e. key down
end;

procedure TLemmingsWorld.HandleKeys;
var ks:TKeyboardState;
begin
  ZeroMemory(@ks,sizeof(ks));
  GetKeyboardState(ks);
  case FState of
    rsLevelSelect : begin
      if KeyIsPressed(ks[VK_UP]) and (LevelIndex>0) then SetLevelIndex(LevelIndex-1);
      if KeyIsPressed(ks[VK_DOWN]) and (LevelIndex<LevelList.Count-1) then SetLevelIndex(LevelIndex+1);
      if KeyIsPressed(ks[VK_RETURN]) then begin
        LoadLevel('resource\level\'+LevelList[LevelIndex]);
        ChangeState(rsGame);
      end;
    end;
    rsGame : begin
      if KeyIsPressed(ks[VK_DOWN]) and
            (FViewXY.Y<FMap.Level.Height-ScreenVisHeight) then inc(FViewXY.Y,ViewMoveSpeed);
      if KeyIsPressed(ks[VK_UP]) and
            (FViewXY.Y>0) then dec(FViewXY.Y,ViewMoveSpeed);
      if KeyIsPressed(ks[VK_RIGHT]) and
            (FViewXY.X<FMap.Level.Width-ScreenVisWidth) then inc(FViewXY.X,ViewMoveSpeed);
      if KeyIsPressed(ks[VK_Left]) and
            (FViewXY.X>0) then dec(FViewXY.X,ViewMoveSpeed);
    end;
  end;
end;

procedure TLemmingsWorld.SetLevelIndex(new: integer);
var x:TXMLFile;
begin
  if LevelIndex<>New then begin
    LevelIndex:= new;
    x:= TXMLFile.Create;
    try
      x.LoadFromFile('resource\level\'+LevelList[LevelIndex]);
      LevelPrev.LoadFromFile('resource\level\'+x.GetText('level/file'));
//      LevelPrev.RemoveAlpha;
      LevelPrev.Target:= GL_TEXTURE_RECTANGLE_EXT;
      LevelPrev.GenTexture;
      LevelCreator:= x.GetText('level/creator');
      LevelTitle:= x.GetText('level/title');
    finally
      x.Free;
    end;
  end;
end;

function TLemmingsWorld.GetTimer: DWORD;
begin
  Result:= GetTickCount;
end;


procedure TLemmingsWorld.HandleMouseClick(pt: TPoint);
begin
  FControlBar.HandleMouseClick(pt);
end;

{ TLemmingsControlbar }

constructor TLemmingsControlbar.Create(AWorld: TLemmingsWorld);
begin
  inherited;
  Position:= GLHelper.MakeVector(0,ScreenVisHeight,0);
  OpenTexture('resource\img\controlbar.tga',0,0,rect(0,0,ScreenVisWidth,ScreenHeight-ScreenVisHeight));
  World.SubObjects.Remove(Self);
  Buttons[cbBlocker].pos:= rect(0,ScreenHeight-80,50,ScreenHeight);
end;

function TLemmingsControlbar.FindButton(pt: TPoint): TControlButton;
begin
  for Result:= high(TControlButton) downto cbNone do
    if PtInRect(Buttons[Result].pos,pt) then break;
end;

procedure TLemmingsControlbar.HandleMouseClick(pt: TPoint);
var b:TControlButton;
begin
  b:= FindButton(pt);
  if (b<>cbNone) and (b<>FCurrentTool) and (Buttons[b].Count>0) then begin
    FCurrentTool:= b;   
  end;
end;

end.
